////////////////////////////////////////
// CheckoutStation Implementation	    //
//																	  //
////////////////////////////////////////

#ifndef  CHECKOUTSTATION_CLASS
#include "CheckoutStation.h"
#endif

#ifndef  EVENTSERVER_CLASS
#include "C:\scs\utility\EventServer.h"
#endif

#ifndef  COMMONEVENT_CLASS
#include "C:\scs\common\CommonEvent.h"
#endif

#ifndef STD__IOSTREAM
#include <iostream>
#endif

using namespace Station;
using std::cout;
using std::endl;

/* Note: To put the EventHandler class inside individual State class
   Advantage:
    1. To serve like the IF style. i.e, have same OpenSessionEventHandler
       in differnt State class but same name ... get it? ;p
    2. Easier to understand and better structure
*/

// ** State Machine "Local" Class Definition **

// ========================================
// *** CheckoutStationState Declaration ***
// ========================================

// State Base Class   7/29/98
// Abstract Base Class that exists to provide Polymorphism
class CheckoutStationState 
{
  public:
    CheckoutStationState(CheckoutStation & p_station)
      : m_station(p_station) {}
    virtual ~CheckoutStationState() = 0 {}  
    CheckoutStation & GetStation(void) { return m_station; }
  private:
    CheckoutStation & m_station;
};

// ==================================
// *** NoSessionState Declaration ***
// ==================================
class NoSessionState : public CheckoutStationState
{
  public:
    NoSessionState(CheckoutStation & p_station);
    virtual ~NoSessionState();

  private:
    // ** Class to "Handle" state_specific Events
    // Call Back Pattern  8/8/98
    // ===========================================
    // *** OpenSessionEventHandler Declaration ***
    // ===========================================
    class OpenSessionEventHandler : public EventHandler
    {
      public:
        OpenSessionEventHandler(NoSessionState & p_state);
        ~OpenSessionEventHandler();
        virtual void HandleEvent (const Event & p_Event);
      private:
        NoSessionState & m_state;
    };

    // Add more Handler Class as necessary
    // ...

    // Auto EventHandler Registration
    OpenSessionEventHandler * m_OpenSessionEventHandlerPtr;
};


// ====================================
// *** SessionOpenState Declaration ***
// ====================================
class SessionOpenState : public CheckoutStationState
{
  public:
    SessionOpenState(CheckoutStation & p_station);
    virtual ~SessionOpenState();

  private:
    // =============================================
    // *** CancelSessionEventHandler Declaration ***
    // =============================================
    class CancelSessionEventHandler : public EventHandler
    {
      public:
        CancelSessionEventHandler(SessionOpenState & p_state);
        ~CancelSessionEventHandler();
        virtual void HandleEvent (const Event & p_Event);
      private:
      SessionOpenState & m_state;
    };

    // Add more CallBack EventHandler class as necessary
    // ...

    // Auto EventHandler Registration
    CancelSessionEventHandler * m_CancelSessionEventHandlerPtr;
};


// ============================================== 
// *** OpenSessionEventHandler Implementation *** 
// ==============================================
NoSessionState::OpenSessionEventHandler::OpenSessionEventHandler(NoSessionState & p_state)
: m_state(p_state)
{
  // Register
  Reactor::GetInstance().RegisterHandler( typeid(OpenSessionEvent), *this);  
}

NoSessionState::OpenSessionEventHandler::~OpenSessionEventHandler()
{
  // Unregister
  Reactor::GetInstance().RemoveHandler( typeid(OpenSessionEvent), *this);   
}

void NoSessionState::OpenSessionEventHandler::HandleEvent (const Event & p_Event)
{
  // Attach New Session
  m_state.GetStation().AttachNewCheckoutSession();
  // Change of Station State
  m_state.GetStation().SetState( new SessionOpenState(m_state.GetStation()) );
  cout << "OpenSessionEventHandler:: OpenSessionEvent has been handled" << endl;
  cout << "Current state is SessionOpenState\n";
}

// ================================================ 
// *** CancelSessionEventHandler Implementation *** 
// ================================================
SessionOpenState::CancelSessionEventHandler::CancelSessionEventHandler(SessionOpenState & p_state)
: m_state(p_state)
{
  // Register
  Reactor::GetInstance().RegisterHandler( typeid(CancelSessionEvent), *this);
}

SessionOpenState::CancelSessionEventHandler::~CancelSessionEventHandler()
{
  // Unregister
  Reactor::GetInstance().RemoveHandler( typeid(CancelSessionEvent), *this); 
}

void SessionOpenState::CancelSessionEventHandler::HandleEvent (const Event & p_Event)
{
  // Detach Session
  m_state.GetStation().DetachCurrentCheckoutSession();
  // Change of Station State
  m_state.GetStation().SetState( new NoSessionState(m_state.GetStation()) );
  cout << "CancelSessionEventHandler:: CancelSessionEvent has been handled." << endl;
  cout << "Current state is NoSessionState\n";
}

// =====================================
// *** NoSessionState Implementation *** 
// =====================================
NoSessionState::NoSessionState(CheckoutStation & p_station)
: CheckoutStationState(p_station)
{
  m_OpenSessionEventHandlerPtr = new OpenSessionEventHandler(*this);
}

NoSessionState::~NoSessionState()
{
  delete m_OpenSessionEventHandlerPtr;
}

// =======================================
// *** SessionOpenState Implementation *** 
// =======================================
SessionOpenState::SessionOpenState(CheckoutStation & p_station)
: CheckoutStationState(p_station)
{
  m_CancelSessionEventHandlerPtr = new CancelSessionEventHandler(*this);
}

SessionOpenState::~SessionOpenState()
{
  delete m_CancelSessionEventHandlerPtr;
}

// ===================================
// *** Station Class Implemenation ***
// ===================================

// default ctor
CheckoutStation::CheckoutStation() throw()
  : m_CheckoutSessionPtr(0), // initialization
    m_Name("Unknown Station")
{ 
  // pass in *this == Station to NoSessionState  
  m_CheckoutStationStatePtr = new NoSessionState(*this);
}

CheckoutStation::CheckoutStation(Common::String Name) throw()
  : m_CheckoutSessionPtr(0), m_Name(Name)
{
  m_CheckoutStationStatePtr = new NoSessionState(*this);
}

// dtor
CheckoutStation::~CheckoutStation()
{
  if ( m_CheckoutSessionPtr != 0 )
    delete m_CheckoutSessionPtr;
  if ( m_CheckoutStationStatePtr != 0 )
    delete m_CheckoutStationStatePtr;
}

// State Machine 7/29/98
void CheckoutStation::SetState(CheckoutStationState * p_NewStatePtr)
{
  if ( m_CheckoutStationStatePtr != 0 )
    delete m_CheckoutStationStatePtr;
  m_CheckoutStationStatePtr = p_NewStatePtr;
}

// singleton GetInstance
CheckoutStation & CheckoutStation::GetInstance(void)
{
  static CheckoutStation as_Instance;
  return as_Instance;
}

void CheckoutStation::AttachNewCheckoutSession() throw(Common::InternalException)
{
  this->DetachCurrentCheckoutSession();
  try 
  {
    m_CheckoutSessionPtr = new Sessions::CheckoutSession(*this);
  }
  catch ( std::bad_alloc & p_Exception )
  {
    m_CheckoutSessionPtr = 0;
    throw Common::InternalException( p_Exception.what() );
  }
}

void CheckoutStation::DetachCurrentCheckoutSession() throw()
{
  if ( m_CheckoutSessionPtr != 0 )
  {
    delete m_CheckoutSessionPtr;
    m_CheckoutSessionPtr = 0;
  }
}

// Get function
Sessions::CheckoutSession & CheckoutStation::GetCheckoutSession(void) const throw(NoAttachedSessionException)
{
  if (m_CheckoutSessionPtr == 0)
    throw NoAttachedSessionException();

  return *m_CheckoutSessionPtr;
}

// Get StationName function
Common::String CheckoutStation::GetStationName(void) const throw()
{
  return m_Name;
}